# Processes.py - Provides utility classes for working with processes.
#
# Copyright (C) 2007 Jason J. Roberts
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (available in the file LICENSE.TXT)
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import codecs
import datetime
import locale
import os
import re
import subprocess
import sys
import time
import types

# Try to import the threading and Queue modules, so we can use them in the
# implementation of ChildProcess.Execute(). These modules will be present on
# platforms that support multithreading. If they are not available, we have a
# fallback implementation that gets the job done but not as nicely for the user.

try:
    import threading
    import Queue
except:
    pass

from GeoEco.DynamicDocString import DynamicDocString
from GeoEco.Internationalization import _
from GeoEco.Logging import Logger


class ChildProcess(object):
    __doc__ = DynamicDocString()
    
    def __init__(self, program=None, arguments=None, workingDirectory=None, stdinString=None, stdoutLogLevel=u'Debug', stderrLogLevel=u'Error', universalNewlines=True, failOnNonZeroExit=True, failOnStderr=True, captureBufferSize=1, windowState=None, maxRunTime=None):
        self.__doc__.Obj.ValidateMethodInvocation()

        # Initialize properties from constructor parameters.
        
        self._Program = program
        self._Arguments = arguments
        self._WorkingDirectory = workingDirectory
        self._StdinString = stdinString
        self._StdoutLogLevel = stdoutLogLevel
        self._StderrLogLevel = stderrLogLevel
        self._UniversalNewlines = universalNewlines
        self._FailOnNonZeroExit = failOnNonZeroExit
        self._FailOnStderr = failOnStderr
        self._CaptureBufferSize = captureBufferSize
        self._WindowState = windowState
        self._MaxRunTime = maxRunTime

        # Initialize read-only properties not given by constructor parameters.

        self._StdoutTranscript = ''
        self._StderrTranscript = ''
        self._ExitCode = 0

    def _GetProgram(self):
        return self._Program
    
    def _SetProgram(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._Program = value

    Program = property(_GetProgram, _SetProgram, doc=DynamicDocString())

    def _GetArguments(self):
        return self._Arguments
    
    def _SetArguments(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._Arguments = value

    Arguments = property(_GetArguments, _SetArguments, doc=DynamicDocString())

    def _GetWorkingDirectory(self):
        return self._WorkingDirectory
    
    def _SetWorkingDirectory(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._WorkingDirectory = value

    WorkingDirectory = property(_GetWorkingDirectory, _SetWorkingDirectory, doc=DynamicDocString())

    def _GetStdinString(self):
        if self._StdinString is None:
            return None
        return unicode(self._StdinString, errors='replace')
    
    def _SetStdinString(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._StdinString = value

    StdinString = property(_GetStdinString, _SetStdinString, doc=DynamicDocString())

    def _GetStdinString8Bit(self):
        if self._StdinString is None:
            return None
        return str(self._StdinString)
    
    def _SetStdinString8Bit(self, value):
        if not isinstance(value, (types.StringType, types.NoneType)):
            Logger.RaiseException(TypeError(_(u'The StdinString8Bit property may be set to an 8-bit string (an instance of the Python types.StringType type) or None.')))
        self._StdinString = value

    StdinString8Bit = property(_GetStdinString8Bit, _SetStdinString8Bit, doc=DynamicDocString())

    def _GetStdoutLogLevel(self):
        return self._StdoutLogLevel
    
    def _SetStdoutLogLevel(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._StdoutLogLevel = value

    StdoutLogLevel = property(_GetStdoutLogLevel, _SetStdoutLogLevel, doc=DynamicDocString())

    def _GetStderrLogLevel(self):
        return self._StderrLogLevel
    
    def _SetStderrLogLevel(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._StderrLogLevel = value

    StderrLogLevel = property(_GetStderrLogLevel, _SetStderrLogLevel, doc=DynamicDocString())

    def _GetUniversalNewlines(self):
        return self._UniversalNewlines
    
    def _SetUniversalNewlines(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._UniversalNewlines = value

    UniversalNewlines = property(_GetUniversalNewlines, _SetUniversalNewlines, doc=DynamicDocString())

    def _GetFailOnNonZeroExit(self):
        return self._FailOnNonZeroExit
    
    def _SetFailOnNonZeroExit(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._FailOnNonZeroExit = value

    FailOnNonZeroExit = property(_GetFailOnNonZeroExit, _SetFailOnNonZeroExit, doc=DynamicDocString())

    def _GetFailOnStderr(self):
        return self._FailOnStderr
    
    def _SetFailOnStderr(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._FailOnStderr = value

    FailOnStderr = property(_GetFailOnStderr, _SetFailOnStderr, doc=DynamicDocString())

    def _GetCaptureBufferSize(self):
        return self._CaptureBufferSize
    
    def _SetCaptureBufferSize(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._CaptureBufferSize = value

    CaptureBufferSize = property(_GetCaptureBufferSize, _SetCaptureBufferSize, doc=DynamicDocString())

    def _GetWindowState(self):
        return self._WindowState
    
    def _SetWindowState(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._WindowState = value

    WindowState = property(_GetWindowState, _SetWindowState, doc=DynamicDocString())

    def _GetMaxRunTime(self):
        return self._MaxRunTime
    
    def _SetMaxRunTime(self, value):
        self.__doc__.Obj.ValidatePropertyAssignment()
        self._MaxRunTime = value

    MaxRunTime = property(_GetMaxRunTime, _SetMaxRunTime, doc=DynamicDocString())

    # Read-only properties.    

    def _GetStdoutTranscript(self):
        if self._StdoutTranscript is None:
            return None
        return unicode(self._StdoutTranscript, errors='replace')

    StdoutTranscript = property(_GetStdoutTranscript, doc=DynamicDocString())

    def _GetStdoutTranscript8Bit(self):
        return self._StdoutTranscript

    StdoutTranscript8Bit = property(_GetStdoutTranscript8Bit, doc=DynamicDocString())

    def _GetStderrTranscript(self):
        if self._StderrTranscript is None:
            return None
        return unicode(self._StderrTranscript, errors='replace')

    StderrTranscript = property(_GetStderrTranscript, doc=DynamicDocString())

    def _GetStderrTranscript8Bit(self):
        return self._StderrTranscript

    StderrTranscript8Bit = property(_GetStderrTranscript8Bit, doc=DynamicDocString())

    def _GetExitCode(self):
        return self._ExitCode

    ExitCode = property(_GetExitCode, doc=DynamicDocString())

    # Event codes used for cross-thread communication.

    _EVENT_STDOUT_LINE_RECEIVED = 1
    _EVENT_STDERR_LINE_RECEIVED = 2
    _EVENT_STDIN_CLOSED = 3
    _EVENT_STDOUT_CLOSED = 4
    _EVENT_STDERR_CLOSED = 5

    # Worker threads used for I/O with the child process.

    def _WriterThread(self, stdin, stdinString, eventQueue):
        stdin.write(stdinString)
        stdin.close()
        eventQueue.put((ChildProcess._EVENT_STDIN_CLOSED, None))

    def _ReaderThread(self, handle, captureBufferSize, lineReceivedEvent, handleClosedEvent, eventQueue):
        currentLine = ''
        data = handle.read(captureBufferSize)
        while data != '':
            currentLine += data
            lines = currentLine.split('\n')
            for line in lines[:-1]:
                eventQueue.put((lineReceivedEvent, line + '\n'))
            currentLine = lines[-1]
            data = handle.read(captureBufferSize)
        if currentLine != '':
            eventQueue.put((lineReceivedEvent, currentLine))
        handle.close()
        eventQueue.put((handleClosedEvent, None))

    def Execute(self):
        self.__doc__.Obj.ValidateMethodInvocation()

        # Fail if the caller did not set self.Program yet.

        if self.Program is None:
            Logger.RaiseException(TypeError(_(u'The Program property must be set before Execute is invoked.')))
            
        try:

            # If we are running on MS Windows, initialize the startupinfo
            # structure and creationflags parameter.

            startupinfo = None
            creationflags = 0

            if sys.platform.lower() == u'win32':

                # Sadly, Python's subprocess module does not actually use the
                # pywin32 modules to create the process (unless you modify the
                # subprocess module after Python is installed). The subprocess
                # module defines its own version of STARTUPINFO, which only has
                # a small subset of the options supported by Microsoft's
                # STARTUPINFO:
                #
                #     class STARTUPINFO:
                #        dwFlags = 0
                #        hStdInput = None
                #        hStdOutput = None
                #        hStdError = None
                #        wShowWindow = 0
                #
                # As a result, we can only provide our callers with what is
                # available here. Since we're managing the stream handles, the
                # only thing left to provide is wShowWindow.
                #
                # I can guess why this decision was made--to not create a
                # dependency in the base Python distribution on a Windows-
                # specific library--but it would be nice if they provided full
                # support for STARTUPINFO. The win32 version of GeoEco already
                # has a dependency on pywin32, but in the spirit of minimizing
                # that dependency, since subprocess already worked so hard to
                # minimize it, I hardcode some needed constants rather than
                # importing the win32process and win32con modules.

                if self.WindowState is not None:
                    startupinfo = subprocess.STARTUPINFO()
                    if self.WindowState == u'invisible':
                        creationflags = 0x08000000                  # win32process.CREATE_NO_WINDOW
                    elif self.WindowState == u'maximized':
                        startupinfo.dwFlags |= 0x00000001       # win32process.STARTF_USESHOWWINDOW
                        startupinfo.wShowWindow = 3             # win32con.SW_SHOWMAXIMIZED
                    elif self.WindowState == u'minimized':
                        startupinfo.dwFlags |= 0x00000001       # win32process.STARTF_USESHOWWINDOW
                        startupinfo.wShowWindow = 7             # win32con.SW_SHOWMINNOACTIVE

            # Start the program.
            #
            # Note: subprocess.Popen does not appear to support
            # Unicode. When we pass Unicode to it, it tries to encode
            # it using the ascii codec. If there are any characters
            # above ordinal 127, this will fail. But occasionally the
            # caller might provide a character in that range,
            # particularly when passing file names on a non-English
            # operating system. To maximize chances for success, first
            # encode the Unicode in the locale's preferred encoding
            # before calling subprocess.Popen.

            programName = os.path.basename(self.Program)

            args = [self.Program]
            if self.Arguments is not None:
                args.extend(self.Arguments)

            encoder = codecs.getencoder(locale.getpreferredencoding())
            for i in range(len(args)):
                if isinstance(args[i], unicode):
                    args[i] = encoder(args[i])[0]

            Logger.Debug(_(u'Executing program: %s'), subprocess.list2cmdline(args))

            p = subprocess.Popen(args=args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=self.WorkingDirectory, universal_newlines=self.UniversalNewlines, startupinfo=startupinfo, creationflags=creationflags)

            # If we're running on a multithreaded platform, rather than writing
            # stdin and reading stdout and stderr from this thread, we will
            # create a new thread for each of those tasks. This design is
            # similar to the internal implementation of Popen.communicate()
            # and provides several advantages vs. calling communicate() or
            # writing our own poll-and-sleep loop:
            #
            # 1. When the program exits, we will detect it immediately.
            #
            # 2. It eliminates the chance that a deadlock would occur in which
            #    we are in a blocking write to stdin while the child process is
            #    in a blocking write to stdout or stderr. (I have not observed
            #    this deadlock, but in theory it could occur depending on the
            #    stream and pipe implementations in Python and the child
            #    process.)
            #
            # 3. Output from the program will be logged as soon as it can be
            #    read from its stdout and stderr. This gives feedback quickly
            #    to the user and also increases the chances that stdout and
            #    stderr messages will be ordered properly. (Of course, these
            #    benefits will be undermined if the child process or Python
            #    buffers the stdout and stderr streams.)
            #
            # Unfortunately, I doubted the threads could successfully write log
            # messages. The Python logging module appears more-or-less thread
            # safe, but I don't think the ArcGIS geoprocessor IDispatch COM
            # object can be successfully invoked from another thread, unless we
            # do the tedious COM work to deal with threads and COM "apartments"
            # properly. (And, even then, the IDispatch object itself might not
            # be thread safe.) I have not had good luck creating more than one
            # geoprocessor IDispatch object before. So, rather than prototyping
            # those very complicated approaches just so the threads can do
            # logging directly, the threads send the captured stdout and stderr
            # text back to the main thread using a Python queue.

            if sys.modules.has_key(u'threading') and sys.modules.has_key(u'Queue'):

                # Initialize variables used to manage inter-thread
                # communication.

                stdinClosed = False
                stdoutClosed = False
                stderrClosed = False
                eventQueue = Queue.Queue()
                stdoutLines = []
                stderrLines = []

                # Start the threads.

                if self._StdinString is not None:
                    stdinThread = threading.Thread(target=self._WriterThread, args=(p.stdin, self._StdinString, eventQueue))
                    stdinThread.setDaemon(True)            # Copied this from subprocess module. Not sure why it is done. Maybe so process can be killed easily on UNIX without kill -9?
                    stdinThread.start()
                else:
                    stdinThread = None
                    p.stdin.close()
                    stdinClosed = True

                stdoutThread = threading.Thread(target=self._ReaderThread, args=(p.stdout, self.CaptureBufferSize, ChildProcess._EVENT_STDOUT_LINE_RECEIVED, ChildProcess._EVENT_STDOUT_CLOSED, eventQueue))
                stdoutThread.setDaemon(True)
                stdoutThread.start()

                stderrThread = threading.Thread(target=self._ReaderThread, args=(p.stderr, self.CaptureBufferSize, ChildProcess._EVENT_STDERR_LINE_RECEIVED, ChildProcess._EVENT_STDERR_CLOSED, eventQueue))
                stderrThread.setDaemon(True)
                stderrThread.start()

                # Loop until the threads tell us that all three streams are
                # closed.

                startTime = datetime.datetime.now()

                while not stdinClosed or not stdoutClosed or not stderrClosed:

                    # Wait until we get a message from one of the threads, or
                    # the maximum run time elapses.

                    if self.MaxRunTime is None:
                        (event, line) = eventQueue.get()
                    else:
                        timeRemaining = startTime - datetime.datetime.now() + datetime.timedelta(seconds=self.MaxRunTime)
                        secondsRemaining = timeRemaining.days * 86400 + timeRemaining.seconds
                        if secondsRemaining <= 0:
                            secondsRemaining = 1
                        try:
                            (event, line) = eventQueue.get(timeout=secondsRemaining)
                        except Queue.Empty:
                            Logger.Error(_(u'%(program)s has exceeded the maximum allowed run time of %(limit)i seconds.') % {u'program' : programName, u'limit' : self.MaxRunTime})
                            self._Kill(p)

                    # Process all queued messages sent by the threads.

                    while True:

                        # Log the line of output (if any) and append it to the
                        # appropriate buffer.

                        if event == ChildProcess._EVENT_STDOUT_LINE_RECEIVED:
                            stdoutLines.append(line)
                            if self.StdoutLogLevel is not None:
                                self._LogReceivedMessage(programName, unicode(line, errors='replace').rstrip(), self.StdoutLogLevel)

                        elif event == ChildProcess._EVENT_STDERR_LINE_RECEIVED:
                            stderrLines.append(line)
                            if self.StderrLogLevel is not None:
                                self._LogReceivedMessage(programName, unicode(line, errors='replace').rstrip(), self.StderrLogLevel)

                        # If a handle was closed, take note of it.                            

                        elif event == ChildProcess._EVENT_STDIN_CLOSED:
                            stdinClosed = True
                            
                        elif event == ChildProcess._EVENT_STDOUT_CLOSED:
                            stdoutClosed = True
                            
                        elif event == ChildProcess._EVENT_STDERR_CLOSED:
                            stderrClosed = True

                        else:
                            Logger.RaiseException(AssertionError(u'Received an invalid event (%i) from eventQueue' % event))

                        # Get the next message from the queue.

                        try:
                            (event, line) = eventQueue.get(block=False)
                        except Queue.Empty:
                            break

                self._StdoutTranscript = ''.join(stdoutLines)
                self._StderrTranscript = ''.join(stderrLines)
                        
                # Wait for the threads to exit.

                if stdinThread is not None:
                    stdinThread.join()
                stdoutThread.join()
                stderrThread.join()

            # If we are not running on a multithreaded platform, use
            # Popen.communicate. This will immediately detect that the program
            # has exited, but does not provide the other two advantages of the
            # multithreaded approach (these are discussed above).

            else:            
                (self._StdoutTranscript, self._StderrTranscript) = p.communicate(self._StdinString)
                
                for line in self.StdoutTranscript.splitlines():
                    self._LogReceivedMessage(programName, line, self.StdoutLogLevel)
                    
                for line in self.StderrTranscript.splitlines():
                    self._LogReceivedMessage(programName, line, self.StderrLogLevel)

            # Capture the process exit code.
            
            self._ExitCode = p.wait()

            # If requested by the caller, fail if the process exited with a non-
            # zero exit code or output anything to the stderr stream.

            if self.FailOnNonZeroExit and self.ExitCode != 0:
                Logger.RaiseException(RuntimeError(_(u'%(name)s returned exit code %(code)i, indicating failure.') % {u'name' : programName, u'code' : self.ExitCode}))
            else:
                Logger.Debug(_(u'%(name)s returned exit code %(code)i.') % {u'name' : programName, u'code' : self.ExitCode})

            if self.FailOnStderr and len(self.StderrTranscript) > 0:
                Logger.RaiseException(RuntimeError(_(u'%s failed. Please review the preceding log messages for more information.') % programName))

        except:
            Logger.LogExceptionAsError(_(u'Execution of %s failed.') % self.Program)
            raise

    def _Kill(self, p):
        assert isinstance(p, subprocess.Popen), u'p must be an instance of subprocess.Popen'

        programName = os.path.basename(self.Program)        

        # On Windows, use Win32 TerminateProcess to kill the process.
        
        if sys.platform.lower() == u'win32':
            import win32api, win32con
            Logger.Debug(_(u'Calling Win32 TerminateProcess() on process %i.') % p.pid)
            process = win32api.OpenProcess(win32con.PROCESS_TERMINATE, False, p.pid)
            win32api.TerminateProcess(process, 1)
            Logger.RaiseException(RuntimeError(_(u'%(program)s (process ID %(pid)i) was terminated because it did not complete its execution within the maximum allowed run time.') % {u'program' : programName, u'pid' : p.pid}))

        # On other platforms, try to use kill. WARNING: This code has not been
        # tested yet.

        elif hasattr(os, u'kill'):
            for sig in [15, 2, 9]:      # Try these signals, in this order: SIGTERM, SIGINT, SIGKILL
                Logger.Debug(_(u'Sending signal %(sig)i to process %(pid)i.') % {u'sig' : sig, u'pid' : p.pid})
                os.kill(p.pid, sig)
                for i in range(60):
                    time.sleep(1)
                    if p.poll() is not None:
                        Logger.RaiseException(RuntimeError(_(u'%(program)s (process ID %(pid)i) was killed because it did not complete its execution within the maximum allowed run time.') % {u'program' : programName, u'pid' : p.pid}))

        # Kill is not available.            
                        
        Logger.RaiseException(RuntimeError(_(u'%(program)s (process ID %(pid)i) exceeded the maximum allowed run time but could not be killed because this platform does not implement the os.kill() function.') % {u'program' : programName, u'pid' : p.pid}))

    def _LogReceivedMessage(self, programName, message, logLevel):

        # Ugly hack: Sometimes we run Python programs that initialize
        # the GeoEco logging system and use it to report messages. By
        # default, these messages are formatted with the date, time
        # and log level. If we find that, strip it out.

        m = re.match('^\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d[,\.]\d\d\d (DEBUG|INFO|WARNING|ERROR) (.*)$', message)
        if m is not None:
            logLevel = m.groups()[0].lower()
            message = m.groups()[1]
        else:
            message = '%s: %s' % (programName, message)

        # Log the message.
        
        if logLevel == u'debug':
            Logger.Debug(message)
        elif logLevel == u'info':
            Logger.Info(message)
        elif logLevel == u'warning':
            Logger.Warning(message)
        elif logLevel == u'error':
            Logger.Error(message)
        else:
            Logger.RaiseException(AssertionError(u'Invalid logLevel (%s)' % logLevel))

    @classmethod
    def GetBinDirectory(cls):
        import GeoEco
        return unicode(os.path.join(os.path.dirname(sys.modules[u'GeoEco'].__file__), 'Bin', sys.platform.lower()))

    @classmethod
    def ExecuteProgram(cls, program, arguments=None, workingDirectory=None, stdinString=None, stdoutLogLevel=u'Debug', stderrLogLevel=u'Error', universalNewlines=True, failOnNonZeroExit=True, failOnStderr=True, captureBufferSize=1, windowState=None, maxRunTime=None):
        cls.__doc__.Obj.ValidateMethodInvocation()
        process = ChildProcess(program=program, arguments=arguments, workingDirectory=workingDirectory, stdinString=stdinString, stdoutLogLevel=stdoutLogLevel, stderrLogLevel=stderrLogLevel, universalNewlines=universalNewlines, failOnNonZeroExit=failOnNonZeroExit, failOnStderr=failOnStderr, captureBufferSize=captureBufferSize, windowState=windowState, maxRunTime=maxRunTime)
        process.Execute()
        return (process.ExitCode, process.StdoutTranscript, process.StderrTranscript)

    @classmethod
    def ExecuteProgramInBinDirectory(cls, programName, arguments=None, workingDirectory=None, stdinString=None, stdoutLogLevel=u'Debug', stderrLogLevel=u'Error', universalNewlines=True, failOnNonZeroExit=True, failOnStderr=True, captureBufferSize=1, windowState=None, maxRunTime=None):
        cls.__doc__.Obj.ValidateMethodInvocation()
        process = ChildProcess(arguments=arguments, workingDirectory=workingDirectory, stdinString=stdinString, stdoutLogLevel=stdoutLogLevel, stderrLogLevel=stderrLogLevel, universalNewlines=universalNewlines, failOnNonZeroExit=failOnNonZeroExit, failOnStderr=failOnStderr, captureBufferSize=captureBufferSize, windowState=windowState, maxRunTime=maxRunTime)
        program = os.path.join(cls.GetBinDirectory(), programName)
        if sys.platform.lower() == u'win32' and len(os.path.splitext(program)[1]) <= 0 and not os.path.isfile(program):
            if os.path.isfile(program + u'.exe'):
                program += u'.exe'
            elif os.path.isfile(program + u'.cmd'):
                program += u'.cmd'
            elif os.path.isfile(program + u'.bat'):
                program += u'.bat'
            elif os.path.isfile(program + u'.com'):
                program += u'.com'
        process.Program = program
        process.Execute()
        return (process.ExitCode, process.StdoutTranscript, process.StderrTranscript)

    @classmethod
    def GetPythonExecutable(cls):
        cls.__doc__.Obj.ValidateMethodInvocation()
        
        Logger.Debug(_(u'Determining the path to the Python executable.'))

        if sys.platform == 'win32':

            # First check for python.exe in the Python "prefix"
            # directory for the currently executing Python
            # interpreter.

            if sys.prefix is not None and os.path.isfile(os.path.join(sys.prefix, 'python.exe')):
                executable = os.path.join(sys.prefix, 'python.exe')
                Logger.Debug(_(u'Using the path %(path)s, obtained from sys.prefix' % {u'path': executable}))

            # Otherwise ask Windows for the program that handles
            # Python files.

            else:
                Logger.Debug(_(u'Unable to find python.exe using sys.prefix'))

                import win32api

                try:
                    executable = win32api.FindExecutable(__file__, os.path.dirname(__file__))[1]
                except Exception, e:
                    Logger.Debug(_(u'win32api.FindExecutable(%(file)r, %(dir)r) failed with %(e)s: %(msg)s' % {u'file': __file__, u'dir': os.path.dirname(__file__), u'e': e.__class__.__name__, u'msg': e}))
                    import _winreg
                    h = _winreg.OpenKey(_winreg.HKEY_CLASSES_ROOT, 'Python.File\\shell\\open\\command')
                    try:
                        try:
                            executable = _winreg.QueryValue(h, None).split()[0].strip().strip('"').strip("'")
                        except Exception, e:
                            raise RuntimeError(_(u'Unable to determine the path to the Python executable. win32 FindExecutable failed and HKEY_CLASSES_ROOT\\Python.File\\shell\\open\\command could not be read from the registry due to %(e)s: %(msg)s. Check that Python is properly installed.' % {u'e': e.__class__.__name__, u'msg': e}))
                        Logger.Debug(_(u'HKEY_CLASSES_ROOT\\Python.File\\shell\\open\\command contained the path %(path)r' % {u'path': executable}))
                    finally:
                        _winreg.CloseKey(h)
                else:
                    Logger.Debug(_(u'win32api.FindExecutable returned %(path)r' % {u'path': executable}))
                
            return executable

        raise NotImplementedError(_('ChildProcess.GetPythonExecutable is not impelemented for this platform.'))


###############################################################################
# Metadata: module
###############################################################################

from GeoEco.Metadata import *
from GeoEco.Types import *

AddModuleMetadata(shortDescription=_(u'Provides utility classes for working with processes.'))

###############################################################################
# Metadata: ChildProcess class
###############################################################################

AddClassMetadata(ChildProcess,
    shortDescription=_(u'Provides methods for starting a child process and capturing its exit code and output.'),
    isExposedAsCOMServer=True,
    comIID=u'{9C090A6C-9ECB-4BEA-8E96-662407295580}',
    comCLSID=u'{46242EF3-BD03-42E0-8D83-C3384FF4AE22}')

# Public properties

AddPropertyMetadata(ChildProcess.Program,
    typeMetadata=FileTypeMetadata(canBeNone=True),
    shortDescription=_(u'Path to the program to execute, or the name of a program that is reachable through the PATH environment variable.'),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.Arguments,
    typeMetadata=ListTypeMetadata(elementType=UnicodeStringTypeMetadata(stripWhitespace=False), canBeNone=True),
    shortDescription=_(u'List of strings to pass to the program as command-line arguments.'),
    longDescription=_(
u"""The strings are stripped of leading and trailing whitespace, concatenated
together, and passed to the program as a single string when it is executed. As
per programming tradition, the program is responsible for parsing the string
into individual arguments.

The concatenation proceeds according to the rules of the Microsoft C compiler,
as interpreted by the Python subprocess.list2cmdline function. In Python 2.4.4,
the rules were:

1. Arguments are separated by spaces (one space between each argument).

2. For each quotation mark (") that appears in an argument, a backslash (\\) is
   inserted before it.

3. When a series of one or more backslashes precedes a quotation mark (not
   counting the backslashes inserted by rule 2), a backslash is inserted before
   each backslash in the series. (Effectively, the series is doubled in length.)

4. Quotation marks (") are placed around arguments that contain spaces or tabs.

For example, these arguments::

    42
    Hello, World!
    \\
    \\\\
    " \\ \\" \\\\" \\\\\\"

would be concatenated into::

    42 "Hello, World!" \\ \\\\ "\\" \\ \\\\\\" \\\\\\\\\\" \\\\\\\\\\\\\\""
"""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.WorkingDirectory,
    typeMetadata=DirectoryTypeMetadata(mustExist=True, canBeNone=True),
    shortDescription=_(u"""Working directory to start the program in."""),
    longDescription=_(
u"""If a value is not provided, the program is started in the working directory
of the process that is hosting the Python interpreter."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.StdinString,
    typeMetadata=UnicodeStringTypeMetadata(canBeNone=True),
    shortDescription=_(
u"""Input text to pass to the program through the STDIN stream (not through the
command line arguments)."""),
    longDescription=_(
u"""Programs that do not have graphical user interfaces usually interact with
the user through a scrolling text window, such as the "DOS prompt" on Windows
or a "shell" on UNIX operating systems. Data is exchanged between the window
and the program through three "streams". The window sends input data to the
program through the STDIN stream, and the program sends output data back
through the STDOUT stream and error messages through the STDERR stream.
(Some programs send everything over STDOUT and nothing over STDERR.) This Python
script takes the place of the scrolling text window and interacts with the
program by manipulating the streams.

If a value is provided for this parameter, it will be sent to the program
through the STDIN stream, as if the typed it in interactively. If no value is
provided, nothing will be sent to STDIN."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.StdinString8Bit,
    typeMetadata=ClassInstanceTypeMetadata(cls=types.StringType, canBeNone=True),
    shortDescription=_(
u"""8-bit string to pass to the program through the STDIN stream. This
property is identical to the StdinString property except it is an
8-bit string rather than a Unicode string. Whichever property was set
last will be sent to the program."""),
    isExposedToPythonCallers=True)

AddPropertyMetadata(ChildProcess.StdoutLogLevel,
    typeMetadata=UnicodeStringTypeMetadata(canBeNone=True, makeLowercase=True, allowedValues=[u'Debug', u'Info', u'Warning', u'Error']),
    shortDescription=_(u"""Severity level for logging the program's output captured from the STDOUT stream."""),
    longDescription=_(
u"""Programs that do not have graphical user interfaces usually interact with
the user through a scrolling text window, such as the "DOS prompt" on Windows
or a "shell" on UNIX operating systems. Data is exchanged between the window
and the program through three "streams". The window sends input data to the
program through the STDIN stream, and the program sends output data back
through the STDOUT stream and error messages through the STDERR stream.
(Some programs send everything over STDOUT and nothing over STDERR.) This Python
script takes the place of the scrolling text window and interacts with the
program by manipulating the streams.

If this parameter is set to one of the following strings, the text sent by the
program through the STDOUT stream will be logged at the severity level specified
by the string:

* Debug - (the default) these are filtered out unless you have customized the
  configuration of the logging system to keep them

* Info - these are "normal" messages, indicating successful progress

* Warning - these are flagged for your attention, but do not indicate that
  processing should be halted

* Error - these are flagged prominantly and indicate that processing should be
  or is about to be halted

If no value is provided, STDOUT text will not be not logged."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.StderrLogLevel,
    typeMetadata=UnicodeStringTypeMetadata(canBeNone=True, makeLowercase=True, allowedValues=[u'Debug', u'Info', u'Warning', u'Error']),
    shortDescription=_(u"""Severity level for logging the program's output captured from the STDERR stream."""),
    longDescription=_(
u"""Programs that do not have graphical user interfaces usually interact with
the user through a scrolling text window, such as the "DOS prompt" on Windows
or a "shell" on UNIX operating systems. Data is exchanged between the window
and the program through three "streams". The window sends input data to the
program through the STDIN stream, and the program sends output data back
through the STDOUT stream and error messages through the STDERR stream.
(Some programs send everything over STDOUT and nothing over STDERR.) This Python
script takes the place of the scrolling text window and interacts with the
program by manipulating the streams.

If this parameter is set to one of the following strings, the text sent by the
program through the STDERR stream will be logged at the severity level specified
by the string:

* Debug - these are filtered out unless you have customized the configuration of
  the logging system to keep them

* Info - these are "normal" messages, indicating successful progress

* Warning - these are flagged for your attention, but do not indicate that
  processing should be halted

* Error - (the default) these are flagged prominantly and indicate that
  processing should be or is about to be halted

If no value is provided, STDERR text will not be not logged."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.UniversalNewlines,
    typeMetadata=BooleanTypeMetadata(),
    shortDescription=_(u'If True, each line of captured program output will be terminated with a single newline character.'),
    longDescription=_(
u"""Depending on how the program was written and what operating system it is
running on, output lines may be terminated with a single linefeed character (the
UNIX convention), a single carriage return character (the Macintosh convention),
or a linefeed followed by a carriage return (the Windows convention).

If this parameter is True, the line termination sequences in captured output are
changed to a single linefeed character, which is traditional in many programming
languages. If this parameter is False, the captured output is left unchanged."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.FailOnNonZeroExit,
    typeMetadata=BooleanTypeMetadata(),
    shortDescription=_(u'If True, an error will be reported if the program returns a non-zero exit code.'),
    longDescription=_(
u"""On most modern operating systems, a program returns an integer error code
when it exits. Traditionally, error code 0 indicates the program executed
successfully, while some other code indicates that it failed.

If this parameter is True, a RuntimeError will be raised if the program returns
a non-zero exit code. If it is False, the exit code will be ignored."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.FailOnStderr,
    typeMetadata=BooleanTypeMetadata(),
    shortDescription=_(u'If True, an error will be reported if the program outputs anything to the STDERR stream.'),
    longDescription=_(
u"""Programs that do not have graphical user interfaces usually interact with
the user through a scrolling text window, such as the "DOS prompt" on Windows
or a "shell" on UNIX operating systems. Data is exchanged between the window
and the program through three "streams". The window sends input data to the
program through the STDIN stream, and the program sends output data back
through the STDOUT stream and error messages through the STDERR stream.
(Some programs send everything over STDOUT and nothing over STDERR.) This Python
script takes the place of the scrolling text window and interacts with the
program by manipulating the streams.

If this parameter is True, a RuntimeError will be raised if the program
transmits anything through the STDERR stream. If it is False, the STDERR output
will be ignored."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.CaptureBufferSize,
    typeMetadata=IntegerTypeMetadata(),
    shortDescription=_(u'Number of bytes to be requested each time captured output is read from the program.'),
    longDescription=_(
u"""Programs that do not have graphical user interfaces usually interact with
the user through a scrolling text window, such as the "DOS prompt" on Windows
or a "shell" on UNIX operating systems. Data is exchanged between the window
and the program through three "streams". The window sends input data to the
program through the STDIN stream, and the program sends output data back
through the STDOUT stream and error messages through the STDERR stream.
(Some programs send everything over STDOUT and nothing over STDERR.) This Python
script takes the place of the scrolling text window and interacts with the
program by manipulating the streams.

The Python function used by this script to read the STDOUT and STDERR streams
appears to be a "blocking" function. This means that if the script tries to
read, say, 100 bytes, the function will not return anything to the script until
the program has transmitted 100 bytes or it has exited. The script is "blocked"
waiting for the program to generate enough output data to fulfill the request
for 100 bytes.

This parameter governs how many bytes are requested by the script each time it
attempts to read one of the streams. The default value, 1, means the script will
never be blocked until it has exhausted the data presently available from the
given stream. This allows the script to log the program's output as quickly as
possible, providing the quickest possible feedback to you about what the program
is doing. But it is also very inefficient, using CPU cycles to retrieve just one
byte at a time even when more are available.

If a larger value is chosen, such as 100, the requests will be much more
efficient but the logging may be delayed if the program outputs too few bytes.
For example, if 100 is chosen and the program outputs 50 bytes, goes to sleep
for five minutes, and then outputs 60 more, the original request for 100 bytes
may not be fulfilled until the second output -- a delay of five minutes. I have
not tested this scenario, but it is theoretically possible.

In general, if your program outputs a relatively small amount of data to these
streams or you care very much about having its output logged as soon as
possible, you should leave this parameter set to 1. If your program outputs a
lot of data or you don't care about the data being logged quickly, increase the
value to something larger, such as 256. If the program outputs megabytes of
data very quickly, my testing has shown that a value higher than 1 million may
be required to obtain optimum performance."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.ExitCode,
    typeMetadata=IntegerTypeMetadata(),
    shortDescription=_(u'Exit code returned by the program.'),
    longDescription=_(
u"""On most modern operating systems, a program returns an integer error code
when it exits. Traditionally, error code 0 indicates the program executed
successfully, while some other code indicates that it failed."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.StdoutTranscript,
    typeMetadata=UnicodeStringTypeMetadata(minLength=0),
    shortDescription=_(u'Program output captured from the STDOUT stream.'),
    longDescription=_(
u"""Programs that do not have graphical user interfaces usually interact with
the user through a scrolling text window, such as the "DOS prompt" on Windows
or a "shell" on UNIX operating systems. Data is exchanged between the window
and the program through three "streams". The window sends input data to the
program through the STDIN stream, and the program sends output data back
through the STDOUT stream and error messages through the STDERR stream.
(Some programs send everything over STDOUT and nothing over STDERR.) This Python
script takes the place of the scrolling text window and interacts with the
program by manipulating the streams.

This output value will be a zero-length string if the program did not generate
any output on the STDOUT stream. The output is always captured and returned, even
if logging of it has been disabled."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.StdoutTranscript8Bit,
    typeMetadata=ClassInstanceTypeMetadata(cls=types.StringType),
    shortDescription=_(
u"""Output of the program sent to the STDOUT stream. This property is
identical to the StdoutTranscript property except it is an 8-bit
string rather than a Unicode string."""),
    isExposedToPythonCallers=True)

AddPropertyMetadata(ChildProcess.StderrTranscript,
    typeMetadata=UnicodeStringTypeMetadata(minLength=0),
    shortDescription=_(u'Program output captured from the STDERR stream.'),
    longDescription=_(
u"""Programs that do not have graphical user interfaces usually interact with
the user through a scrolling text window, such as the "DOS prompt" on Windows
or a "shell" on UNIX operating systems. Data is exchanged between the window
and the program through three "streams". The window sends input data to the
program through the STDIN stream, and the program sends output data back
through the STDOUT stream and error messages through the STDERR stream.
(Some programs send everything over STDOUT and nothing over STDERR.) This Python
script takes the place of the scrolling text window and interacts with the
program by manipulating the streams.

This output value will be a zero-length string if the program did not generate
any output on the STDERR stream. The output is always captured and returned, even
if logging of it has been disabled."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.StderrTranscript8Bit,
    typeMetadata=ClassInstanceTypeMetadata(cls=types.StringType),
    shortDescription=_(
u"""Output of the program sent to the STDERR stream. This property is
identical to the StderrTranscript property except it is an 8-bit
string rather than a Unicode string."""),
    isExposedToPythonCallers=True)

AddPropertyMetadata(ChildProcess.WindowState,
    typeMetadata=UnicodeStringTypeMetadata(canBeNone=True, makeLowercase=True, allowedValues=[u'Invisible', u'Maximized', u'Minimized']),
    shortDescription=_(u'Display state of the program\'s window.'),
    longDescription=_(
u"""This parameter is ignored on platforms other than Microsoft Windows.

If this parameter is not specified, programs that are "console applications"
will execute in a scrolling text window that pops up, and graphical user
interface (GUI) application will appear as they normally do. You can override
the default behavior by specifying one of these options:

* Invisible - a window will not be created, if the program is a console
  application. GUI applications ignore this option.

* Maximized - the window will be maximized. Some GUI applications may ignore
  this request and set their own window state.

* Minimized - the window will be minimized. Some GUI applications may ignore
  this request and set their own window state.

Technical details: The Invisible option is implemented by setting the
CREATE_NO_WINDOW process creation flag when Win32 CreateProcess is called. The
Maximized and Minimized options are implemented by passing a STARTUPINFO
structure to CreateProcess with wShowWindow set to SW_SHOWMAXIMIZED or
SW_SHOWMINNOACTIVE."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddPropertyMetadata(ChildProcess.MaxRunTime,
    typeMetadata=IntegerTypeMetadata(canBeNone=True, minValue=1),
    shortDescription=_(u'Maximum time, in seconds, that the program will be allowed to run.'),
    longDescription=_(
u"""If a value is not provided for this parameter, the program is permitted to
run forever. If a value is provided, the program is forcibly stopped after the
specified time elapses and a RuntimeError is raised.

WARNING: This parameter should only be used to prevent a program that is known
to "hang" from blocking processing forever. You should only use this parameter
when you are certain that there will be no bad consequences from forcibly
stopping the program. When the program is stopped, it may not have the
opportunity to properly shut down pending operations. This could cause data
corruption, for example, if the program is writing to a file or database.

This parameter is ignored if the operating system does not support
multithreading. On those operating systems, the program is permitted to run
forever.

Technical details: On Microsoft Windows, Win32 TerminateProcess is invoked on
the program. On UNIX operating systems, SIGTERM is sent ("kill -15"). If the
program does not stop within 60 seconds, SIGINT is sent ("kill -2"). If it does
not stop after a further 60 seconds, SIGKILL is sent ("kill -9")."""),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

# Constructor

AddMethodMetadata(ChildProcess.__init__,
    shortDescription=_(u'Constructs a new %s instance.') % ChildProcess.__name__,
    isExposedToPythonCallers=True)

AddArgumentMetadata(ChildProcess.__init__, u'self',
    typeMetadata=ClassInstanceTypeMetadata(cls=ChildProcess),
    description=_(u'%s instance.') % ChildProcess.__name__)

AddArgumentMetadata(ChildProcess.__init__, u'program',
    typeMetadata=ChildProcess.Program.__doc__.Obj.Type,
    description=ChildProcess.Program.__doc__.Obj.ShortDescription)

AddArgumentMetadata(ChildProcess.__init__, u'arguments',
    typeMetadata=ChildProcess.Arguments.__doc__.Obj.Type,
    description=ChildProcess.Arguments.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.Arguments.__doc__.Obj.LongDescription)

AddArgumentMetadata(ChildProcess.__init__, u'workingDirectory',
    typeMetadata=ChildProcess.WorkingDirectory.__doc__.Obj.Type,
    description=ChildProcess.WorkingDirectory.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.WorkingDirectory.__doc__.Obj.LongDescription)

AddArgumentMetadata(ChildProcess.__init__, u'stdinString',
    typeMetadata=ChildProcess.StdinString.__doc__.Obj.Type,
    description=ChildProcess.StdinString.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.StdinString.__doc__.Obj.LongDescription)

AddArgumentMetadata(ChildProcess.__init__, u'stdoutLogLevel',
    typeMetadata=ChildProcess.StdoutLogLevel.__doc__.Obj.Type,
    description=ChildProcess.StdoutLogLevel.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.StdoutLogLevel.__doc__.Obj.LongDescription)

AddArgumentMetadata(ChildProcess.__init__, u'stderrLogLevel',
    typeMetadata=ChildProcess.StderrLogLevel.__doc__.Obj.Type,
    description=ChildProcess.StderrLogLevel.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.StderrLogLevel.__doc__.Obj.LongDescription)

AddArgumentMetadata(ChildProcess.__init__, u'universalNewlines',
    typeMetadata=ChildProcess.UniversalNewlines.__doc__.Obj.Type,
    description=ChildProcess.UniversalNewlines.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.UniversalNewlines.__doc__.Obj.LongDescription)

AddArgumentMetadata(ChildProcess.__init__, u'failOnNonZeroExit',
    typeMetadata=ChildProcess.FailOnNonZeroExit.__doc__.Obj.Type,
    description=ChildProcess.FailOnNonZeroExit.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.FailOnNonZeroExit.__doc__.Obj.LongDescription)

AddArgumentMetadata(ChildProcess.__init__, u'failOnStderr',
    typeMetadata=ChildProcess.FailOnStderr.__doc__.Obj.Type,
    description=ChildProcess.FailOnStderr.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.FailOnStderr.__doc__.Obj.LongDescription)

AddArgumentMetadata(ChildProcess.__init__, u'captureBufferSize',
    typeMetadata=ChildProcess.CaptureBufferSize.__doc__.Obj.Type,
    description=ChildProcess.CaptureBufferSize.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.CaptureBufferSize.__doc__.Obj.LongDescription)

AddArgumentMetadata(ChildProcess.__init__, u'windowState',
    typeMetadata=ChildProcess.WindowState.__doc__.Obj.Type,
    description=ChildProcess.WindowState.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.WindowState.__doc__.Obj.LongDescription)

AddArgumentMetadata(ChildProcess.__init__, u'maxRunTime',
    typeMetadata=ChildProcess.MaxRunTime.__doc__.Obj.Type,
    description=ChildProcess.MaxRunTime.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.MaxRunTime.__doc__.Obj.LongDescription)

AddResultMetadata(ChildProcess.__init__, u'childProcess',
    typeMetadata=ClassInstanceTypeMetadata(cls=ChildProcess),
    description=_(u'New %s instance.') % ChildProcess.__name__)

# Public method ChildProcess.Execute

AddMethodMetadata(ChildProcess.Execute,
    shortDescription=_(u'Executes the child process and captures its output.'),
    longDescription=_(u'This function does not return until the process has exited.'),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddArgumentMetadata(ChildProcess.Execute, u'self',
    typeMetadata=ClassInstanceTypeMetadata(cls=ChildProcess),
    description=_(u'%s instance.') % ChildProcess.__name__)

# Public method ChildProcess.ExecuteProgram

AddMethodMetadata(ChildProcess.ExecuteProgram,
    shortDescription=_(u'Executes the specified program and captures its output.'),
    longDescription=_(u'This function does not return until the program has exited.'),
    isExposedToPythonCallers=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Execute Program'),
    arcGISToolCategory=_(u'Data Management\\Programs'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'cls',
    typeMetadata=ClassOrClassInstanceTypeMetadata(cls=ChildProcess),
    description=_(u'%s class or an instance of it.') % ChildProcess.__name__)

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'program',
    typeMetadata=FileTypeMetadata(),
    description=ChildProcess.Program.__doc__.Obj.ShortDescription,
    arcGISDisplayName=_(u'Program to execute'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'arguments',
    typeMetadata=ChildProcess.Arguments.__doc__.Obj.Type,
    description=ChildProcess.Arguments.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.Arguments.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'Command line arguments'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'workingDirectory',
    typeMetadata=ChildProcess.WorkingDirectory.__doc__.Obj.Type,
    description=ChildProcess.WorkingDirectory.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.WorkingDirectory.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'Working directory'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'stdinString',
    typeMetadata=ChildProcess.StdinString.__doc__.Obj.Type,
    description=ChildProcess.StdinString.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.StdinString.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'Input text to pass to STDIN'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'stdoutLogLevel',
    typeMetadata=ChildProcess.StdoutLogLevel.__doc__.Obj.Type,
    description=ChildProcess.StdoutLogLevel.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.StdoutLogLevel.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'STDOUT logging level'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'stderrLogLevel',
    typeMetadata=ChildProcess.StderrLogLevel.__doc__.Obj.Type,
    description=ChildProcess.StderrLogLevel.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.StderrLogLevel.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'STDERR logging level'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'universalNewlines',
    typeMetadata=ChildProcess.UniversalNewlines.__doc__.Obj.Type,
    description=ChildProcess.UniversalNewlines.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.UniversalNewlines.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'Use univeral newline translation'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'failOnNonZeroExit',
    typeMetadata=ChildProcess.FailOnNonZeroExit.__doc__.Obj.Type,
    description=ChildProcess.FailOnNonZeroExit.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.FailOnNonZeroExit.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'Fail if the program exits with a non-0 exit code'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'failOnStderr',
    typeMetadata=ChildProcess.FailOnStderr.__doc__.Obj.Type,
    description=ChildProcess.FailOnStderr.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.FailOnStderr.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'Fail if the program reports anything to STDERR'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'captureBufferSize',
    typeMetadata=ChildProcess.CaptureBufferSize.__doc__.Obj.Type,
    description=ChildProcess.CaptureBufferSize.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.CaptureBufferSize.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'STDOUT and STDERR capture buffer size'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'windowState',
    typeMetadata=ChildProcess.WindowState.__doc__.Obj.Type,
    description=ChildProcess.WindowState.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.WindowState.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'Program window state'))

AddArgumentMetadata(ChildProcess.ExecuteProgram, u'maxRunTime',
    typeMetadata=ChildProcess.MaxRunTime.__doc__.Obj.Type,
    description=ChildProcess.MaxRunTime.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.MaxRunTime.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'Maximum run time'))

AddResultMetadata(ChildProcess.ExecuteProgram, u'exitCode',
    typeMetadata=ChildProcess.ExitCode.__doc__.Obj.Type,
    description=ChildProcess.ExitCode.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.ExitCode.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'Exit code'))

AddResultMetadata(ChildProcess.ExecuteProgram, u'stdoutTranscript',
    typeMetadata=ChildProcess.StdoutTranscript.__doc__.Obj.Type,
    description=ChildProcess.StdoutTranscript.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.StdoutTranscript.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'STDOUT transcript'))

AddResultMetadata(ChildProcess.ExecuteProgram, u'stderrTranscript',
    typeMetadata=ChildProcess.StderrTranscript.__doc__.Obj.Type,
    description=ChildProcess.StderrTranscript.__doc__.Obj.ShortDescription + '\n\n' + ChildProcess.StderrTranscript.__doc__.Obj.LongDescription,
    arcGISDisplayName=_(u'STDERR transcript'))

# Public method ChildProcess.ExecuteProgramInBinDirectory

AddMethodMetadata(ChildProcess.ExecuteProgramInBinDirectory,
    shortDescription=_(u'Executes the specified program residing in the GeoEco Bin directory and captures its output.'),
    longDescription=_(
u"""The Bin directory contains the platform-specific executable programs
distributed with the GeoEco Python package.

This function does not return until the program has exited."""),
    isExposedToPythonCallers=True)

AddArgumentMetadata(ChildProcess.ExecuteProgramInBinDirectory, u'cls',
    typeMetadata=ClassOrClassInstanceTypeMetadata(cls=ChildProcess),
    description=_(u'%s class or an instance of it.') % ChildProcess.__name__)

AddArgumentMetadata(ChildProcess.ExecuteProgramInBinDirectory, u'programName',
    typeMetadata=UnicodeStringTypeMetadata(),
    description=_(u'Filename of the program to execute.'))

for i in range(2, len(ChildProcess.ExecuteProgram.__doc__.Obj.Arguments)):
    argMetadata = ChildProcess.ExecuteProgram.__doc__.Obj.Arguments[i]
    AddArgumentMetadata(ChildProcess.ExecuteProgramInBinDirectory, argMetadata.Name, typeMetadata=argMetadata.Type, description=argMetadata.Description)
del i, argMetadata

for i in range(len(ChildProcess.ExecuteProgram.__doc__.Obj.Results)):
    resultMetadata = ChildProcess.ExecuteProgram.__doc__.Obj.Results[i]
    AddResultMetadata(ChildProcess.ExecuteProgramInBinDirectory, resultMetadata.Name, typeMetadata=resultMetadata.Type, description=resultMetadata.Description)
del i, resultMetadata

# Public method ChildProcess.GetBinDirectory

AddMethodMetadata(ChildProcess.GetBinDirectory,
    shortDescription=_(u'Returns the directory that contains the platform-specific executable programs distributed with the GeoEco Python package.'),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

AddArgumentMetadata(ChildProcess.GetBinDirectory, u'cls',
    typeMetadata=ClassOrClassInstanceTypeMetadata(cls=ChildProcess),
    description=_(u'%s class or an instance of it.') % ChildProcess.__name__)

AddResultMetadata(ChildProcess.GetBinDirectory, u'binDirectory',
    typeMetadata=DirectoryTypeMetadata(),
    description=_(u'Directory that contains the platform-specific executable programs distributed with the GeoEco Python package.'))

# Public method ChildProcess.GetPythonExecutable

AddMethodMetadata(ChildProcess.GetPythonExecutable,
    shortDescription=_(u'Returns the full path to the Python executable that is registered as the opener for .py files.'),
    isExposedToPythonCallers=True,
    isExposedByCOM=True)

CopyArgumentMetadata(ChildProcess.GetBinDirectory, u'cls', ChildProcess.GetPythonExecutable, u'cls')

AddResultMetadata(ChildProcess.GetPythonExecutable, u'path',
    typeMetadata=FileTypeMetadata(),
    description=_(u'Full path to the Python executable that is registered as the opener for .py files.'))


###############################################################################
# Names exported by this module
###############################################################################

__all__ = ['ChildProcess']
